from typing import List, Dict, Union, Any
import numpy as np
import numpy.typing as npt
import torch
import math
import iteround
from stable_baselines3 import PPO
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.policy_type import PolicyType
from csle_common.logging.log import Logger
from csle_common.models.ppo_network import PPONetwork


class PPOPolicy(Policy):
    """
    A neural network policy learned with PPO
    """

    def __init__(self, model: Union[None, PPO, PPONetwork], simulation_name: str, save_path: str,
                 player_type: PlayerType, states: List[State], actions: List[Action],
                 experiment_config: ExperimentConfig, avg_R: float):
        """
        Initializes the policy

        :param model: the PPO model
        :param simulation_name: the simulation name
        :param save_path: the path to save the model to
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        """
        super(PPOPolicy, self).__init__(agent_type=AgentType.PPO, player_type=player_type)
        self.model = model
        self.id = -1
        self.simulation_name = simulation_name
        self.save_path = save_path
        self.load()
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.policy_type = PolicyType.PPO

    def action(self, o: Union[List[float], List[int]], deterministic: bool = True) \
            -> Union[int, float, npt.NDArray[Any]]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        if self.model is None:
            raise ValueError("The model is None")
        if isinstance(self.model, PPO):
            a = self.model.predict(np.array(o), deterministic=deterministic)[0]
            try:
                return int(a)
            except Exception:
                return a
        elif isinstance(self.model, PPONetwork):
            action_tensor, _, _, _ = self.model.get_action_and_value(x=torch.tensor(o))
            try:
                return int(action_tensor.item())
            except Exception:
                return action_tensor.item()
        else:
            raise ValueError(f"Model: {self.model} not recognized")

    def probability(self, o: Union[List[float], List[int]], a: int) -> float:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param o: the action
        :return: the probability of the action
        """
        if self.model is None:
            raise ValueError("The model is None")
        if isinstance(self.model, PPO):
            log_prob = self.model.policy.get_distribution(obs=torch.tensor([o]).to(self.model.device)).log_prob(
                torch.tensor(a).to(self.model.device)).item()
        elif isinstance(self.model, PPONetwork):
            _, log_prob, _, _ = self.model.get_action_and_value(x=torch.tensor(o), action=torch.tensor(a))
        else:
            raise ValueError(f"Model: {self.model} not recognized")
        prob = math.exp(log_prob)
        return prob

    def value(self, o: Union[List[float], List[int]]) -> float:
        """
        Gets the value of a given observation, computed by the critic network

        :param o: the observation to get the value of
        :return: V(o)
        """
        if self.model is None:
            raise ValueError("The model is None")
        if isinstance(self.model, PPO):
            return self.model.policy.predict_values(obs=torch.tensor(np.array([o])).to(self.model.device)).item()
        elif isinstance(self.model, PPONetwork):
            return self.model.get_value(x=torch.tensor(o)).item()
        else:
            raise ValueError(f"Model: {self.model} not recognized")

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["save_path"] = self.save_path
        if self.model is not None:
            self.model.save(path=self.save_path)
            d["policy_kwargs"] = {}
            d["policy_kwargs"]["net_arch"] = []
            if isinstance(self.model, PPO):
                d["policy_kwargs"] = self.model.policy_kwargs
        else:
            d["policy_kwargs"] = {}
            d["policy_kwargs"]["net_arch"] = []
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["experiment_config"] = self.experiment_config.to_dict()
        d["agent_type"] = self.agent_type
        d["avg_R"] = self.avg_R
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "PPOPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = PPOPolicy(model=None, simulation_name=d["simulation_name"], save_path=d["save_path"],
                        states=list(map(lambda x: State.from_dict(x), d["states"])), player_type=d["player_type"],
                        actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                        experiment_config=ExperimentConfig.from_dict(d["experiment_config"]), avg_R=d["avg_R"])
        obj.id = d["id"]
        return obj

    def stage_policy(self, o: Union[List[int], List[float]]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        stage_strategy = np.zeros((len(self.states), len(self.actions)))
        for i, s_a in enumerate(self.states):
            o[0] = s_a
            for j, a in enumerate(self.actions):
                stage_strategy[i][j] = self.probability(o=o, a=j)
            stage_strategy[i] = iteround.saferound(stage_strategy[i], 2)
            assert round(sum(stage_strategy[i]), 3) == 1
        stage_strategy.tolist()
        return list(stage_strategy.tolist())

    def _get_attacker_dist(self, obs) -> List[float]:
        """
        Utility function for getting the action distribution conditioned on a given observation

        :param obs: the observation to condition on
        :return: the conditional ation distribution
        """
        obs = np.array([obs])
        if self.model is None:
            raise ValueError("The model is None")
        if isinstance(self.model, PPO):
            actions, values, log_prob_tensor = self.model.policy.forward(obs=torch.tensor(obs).to(self.model.device))
            log_prob = float(log_prob_tensor.item())
            action = int(actions[0])
        elif isinstance(self.model, PPONetwork):
            action_tensor, log_prob, _, _ = self.model.get_action_and_value(x=torch.tensor(obs))
            action = int(action_tensor.item())
        else:
            raise ValueError(f"Model: {self.model} not recognized")
        if action == 1:
            stop_prob = math.exp(log_prob)
        else:
            stop_prob = 1 - math.exp(log_prob)
        return [1 - stop_prob, stop_prob]

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"model: {self.model}, id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"save path: {self.save_path}, states: {self.states}, experiment_config: {self.experiment_config}," \
               f"avg_R: {self.avg_R}, policy type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "PPOPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return PPOPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "PPOPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())

    def load(self) -> None:
        """
        Attempts to load the policy from disk

        :return: None
        """
        if self.model is None and self.save_path != "":
            try:
                self.model = PPO.load(path=self.save_path)
            except Exception as e1:
                try:
                    self.model = PPONetwork.load(path=self.save_path)
                except Exception as e2:
                    Logger.__call__().get_logger().warning(
                        f"There was an exception loading the model from path: {self.save_path}, "
                        f"exception: {str(e1)}, {repr(e1)} {str(e2)}, {repr(e2)}")



from typing import Optional, List, Dict, Any, Union, Set, Tuple
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_observation.defender.emulation_defender_machine_observation_state \
    import EmulationDefenderMachineObservationState
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_config.kafka_config import KafkaConfig
from csle_common.consumer_threads.docker_stats_consumer_thread import DockerStatsConsumerThread
from csle_common.consumer_threads.aggregated_snort_ids_log_consumer_thread import AggregatedSnortIdsLogConsumerThread
from csle_common.consumer_threads.aggregated_ossec_ids_log_consumer_thread import AggregatedOSSECIdsLogConsumerThread
from csle_common.consumer_threads.aggregated_snort_ids_rule_log_consumer_thread \
    import AggregatedSnortIdsRuleLogConsumerThread
from csle_common.consumer_threads.client_population_consumer_thread import ClientPopulationConsumerThread
from csle_common.consumer_threads.attacker_actions_consumer_thread import AttackerActionsConsumerThread
from csle_common.consumer_threads.defender_actions_consumer_thread import DefenderActionsConsumerThread
from csle_common.consumer_threads.aggregated_host_metrics_thread import AggregatedHostMetricsThread
from csle_collector.client_manager.client_population_metrics import ClientPopulationMetrics
from csle_collector.docker_stats_manager.dao.docker_stats import DockerStats
from csle_collector.snort_ids_manager.dao.snort_ids_alert_counters import SnortIdsAlertCounters
from csle_collector.snort_ids_manager.dao.snort_ids_rule_counters import SnortIdsRuleCounters
from csle_collector.ossec_ids_manager.dao.ossec_ids_alert_counters import OSSECIdsAlertCounters
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_base.json_serializable import JSONSerializable


class EmulationDefenderObservationState(JSONSerializable):
    """
    Represents the defender's agent's current belief state of the emulation
    """

    def __init__(self, kafka_config: Union[KafkaConfig, None], client_population_metrics: ClientPopulationMetrics,
                 docker_stats: DockerStats, snort_ids_alert_counters: SnortIdsAlertCounters,
                 ossec_ids_alert_counters: OSSECIdsAlertCounters, aggregated_host_metrics: HostMetrics,
                 defender_actions: List[EmulationDefenderAction], attacker_actions: List[EmulationAttackerAction],
                 snort_ids_rule_counters: SnortIdsRuleCounters):
        """
        Initializes the DTO

        :param kafka_config: the kafka config
        :param client_population_metrics: the client population metrics
        :param docker_stats: the docker stats
        :param snort_ids_alert_counters: the snort ids alert counters
        :param ossec_ids_alert_counters: the ossec ids alert counters
        :param defender_actions: the list of defender actions
        :param attacker_actions: the list of attacker actions
        :param aggregated_host_metrics: the aggregated host metrics
        :param snort_ids_rule_counters: the aggregated snort IDS rule counters
        """
        self.kafka_config = kafka_config
        self.machines: List[EmulationDefenderMachineObservationState] = []
        self.actions_tried: Set[Tuple[int, int, str]] = set()
        self.client_population_metrics = client_population_metrics
        self.avg_client_population_metrics = self.client_population_metrics.copy()
        self.docker_stats = docker_stats
        self.avg_docker_stats = self.docker_stats.copy()
        self.snort_ids_alert_counters = snort_ids_alert_counters
        self.avg_snort_ids_alert_counters = self.snort_ids_alert_counters.copy()
        self.snort_ids_rule_counters = snort_ids_rule_counters
        self.avg_snort_ids_rule_counters = self.snort_ids_rule_counters.copy()
        self.ossec_ids_alert_counters = ossec_ids_alert_counters
        self.avg_ossec_ids_alert_counters = self.ossec_ids_alert_counters.copy()
        self.attacker_actions = attacker_actions
        self.defender_actions = defender_actions
        self.aggregated_host_metrics = aggregated_host_metrics
        self.avg_aggregated_host_metrics: HostMetrics = self.aggregated_host_metrics.copy()
        self.docker_stats_consumer_thread: Optional[DockerStatsConsumerThread] = None
        self.client_population_consumer_thread: Optional[ClientPopulationConsumerThread] = None
        self.aggregated_snort_ids_log_consumer_thread: Optional[AggregatedSnortIdsLogConsumerThread] = None
        self.aggregated_snort_ids_rule_log_consumer_thread: Optional[AggregatedSnortIdsRuleLogConsumerThread] = None
        self.aggregated_ossec_ids_log_consumer_thread: Optional[AggregatedOSSECIdsLogConsumerThread] = None
        self.attacker_actions_consumer_thread: Optional[AttackerActionsConsumerThread] = None
        self.defender_actions_consumer_thread: Optional[DefenderActionsConsumerThread] = None
        self.aggregated_host_metrics_thread: Optional[AggregatedHostMetricsThread] = None

    def start_monitoring_threads(self) -> None:
        """
        Starts the avg host metrics thread

        :return: None
        """
        if self.kafka_config is None:
            raise ValueError("KafkaConfig is None")
        if self.attacker_actions is None:
            raise ValueError("EmulationAttackerAction is None")
        if self.defender_actions is None:
            raise ValueError("EmulationDefenderAction is None")
        self.aggregated_host_metrics_thread = AggregatedHostMetricsThread(
            host_metrics=self.aggregated_host_metrics,
            sleep_time=self.kafka_config.time_step_len_seconds,
            machines=self.machines
        )
        self.docker_stats_consumer_thread = DockerStatsConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            docker_stats=self.docker_stats)
        self.client_population_consumer_thread = ClientPopulationConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            client_population_metrics=self.client_population_metrics
        )
        self.aggregated_snort_ids_log_consumer_thread = AggregatedSnortIdsLogConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            snort_ids_alert_counters=self.snort_ids_alert_counters
        )
        self.aggregated_snort_ids_rule_log_consumer_thread = AggregatedSnortIdsRuleLogConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            snort_ids_rule_counters=self.snort_ids_rule_counters
        )
        self.aggregated_ossec_ids_log_consumer_thread = AggregatedOSSECIdsLogConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            ossec_ids_alert_counters=self.ossec_ids_alert_counters
        )
        self.attacker_actions_consumer_thread = AttackerActionsConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            attacker_actions=self.attacker_actions
        )
        self.defender_actions_consumer_thread = DefenderActionsConsumerThread(
            kafka_server_ip=self.kafka_config.container.docker_gw_bridge_ip,
            kafka_port=self.kafka_config.kafka_port_external,
            defender_actions=self.defender_actions
        )
        self.aggregated_host_metrics_thread.start()
        self.docker_stats_consumer_thread.start()
        self.client_population_consumer_thread.start()
        self.aggregated_snort_ids_log_consumer_thread.start()
        self.aggregated_snort_ids_rule_log_consumer_thread.start()
        self.aggregated_ossec_ids_log_consumer_thread.start()
        self.attacker_actions_consumer_thread.start()
        self.defender_actions_consumer_thread.start()
        for m in self.machines:
            m.start_monitor_threads()

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationDefenderObservationState":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        kafka_config: Union[None, KafkaConfig] = None
        try:
            kafka_config = KafkaConfig.from_dict(d["kafka_config"])
        except Exception:
            pass
        if "snort_ids_rule_counters" not in d:
            snort_ids_rule_counters = SnortIdsRuleCounters()
        else:
            snort_ids_rule_counters = SnortIdsRuleCounters.from_dict(d["snort_ids_rule_counters"])
        if "avg_snort_ids_rule_counters" not in d:
            avg_snort_ids_rule_counters = SnortIdsRuleCounters()
        else:
            avg_snort_ids_rule_counters = SnortIdsRuleCounters.from_dict(d["avg_snort_ids_rule_counters"])
        obj = EmulationDefenderObservationState(
            kafka_config=kafka_config,
            client_population_metrics=ClientPopulationMetrics.from_dict(d["client_population_metrics"]),
            docker_stats=DockerStats.from_dict(d["docker_stats"]),
            ossec_ids_alert_counters=OSSECIdsAlertCounters.from_dict(d["ossec_ids_alert_counters"]),
            aggregated_host_metrics=HostMetrics.from_dict(d["aggregated_host_metrics"]),
            defender_actions=list(map(lambda x: EmulationDefenderAction.from_dict(x), d["defender_actions"])),
            attacker_actions=list(map(lambda x: EmulationAttackerAction.from_dict(x), d["attacker_actions"])),
            snort_ids_rule_counters=snort_ids_rule_counters,
            snort_ids_alert_counters=SnortIdsAlertCounters.from_dict(d["snort_ids_alert_counters"]))
        obj.machines = list(map(lambda x: EmulationDefenderMachineObservationState.from_dict(d=x), d["machines"]))
        obj.actions_tried = set(list(map(lambda x: (int(x[0]), int(x[1]), str(x[2])), d["actions_tried"])))
        obj.avg_aggregated_host_metrics = HostMetrics.from_dict(d["avg_aggregated_host_metrics"])
        obj.avg_docker_stats = DockerStats.from_dict(d["avg_docker_stats"])
        obj.avg_client_population_metrics = ClientPopulationMetrics.from_dict(d["avg_client_population_metrics"])
        obj.avg_snort_ids_alert_counters = SnortIdsAlertCounters.from_dict(d["avg_snort_ids_alert_counters"])
        obj.avg_snort_ids_rule_counters = avg_snort_ids_rule_counters
        obj.avg_ossec_ids_alert_counters = OSSECIdsAlertCounters.from_dict(d["avg_ossec_ids_alert_counters"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["machines"] = list(map(lambda x: x.to_dict(), self.machines))
        d["actions_tried"] = list(self.actions_tried)
        if self.client_population_metrics is None or self.docker_stats is None \
                or self.snort_ids_alert_counters is None or self.snort_ids_rule_counters is None or \
                self.ossec_ids_alert_counters is None:
            raise ValueError("ClientPopulationMetrics is None")
        d["client_population_metrics"] = self.client_population_metrics.to_dict()
        d["docker_stats"] = self.docker_stats.to_dict()
        d["snort_ids_alert_counters"] = self.snort_ids_alert_counters.to_dict()
        d["snort_ids_rule_counters"] = self.snort_ids_rule_counters.to_dict()
        d["ossec_ids_alert_counters"] = self.ossec_ids_alert_counters.to_dict()
        if self.kafka_config is not None:
            d["kafka_config"] = self.kafka_config.to_dict()
        else:
            d["kafka_config"] = None
        if self.attacker_actions is None:
            raise ValueError("attacker_actions is None and thus has no to_dict attribute")
        if self.defender_actions is None:
            raise ValueError("attacker_actions is None and thus has no to_dict attribute")
        if self.aggregated_host_metrics is None or self.avg_aggregated_host_metrics is None:
            raise ValueError("aggregated_host_metrics is None and thus has no to_dict attribute")
        d["attacker_actions"] = list(map(lambda x: x.to_dict(), self.attacker_actions))
        d["defender_actions"] = list(map(lambda x: x.to_dict(), self.defender_actions))
        d["aggregated_host_metrics"] = self.aggregated_host_metrics.to_dict()
        d["avg_aggregated_host_metrics"] = self.avg_aggregated_host_metrics.to_dict()
        d["avg_client_population_metrics"] = self.avg_client_population_metrics.to_dict()
        d["avg_docker_stats"] = self.avg_docker_stats.to_dict()
        d["avg_snort_ids_alert_counters"] = self.avg_snort_ids_alert_counters.to_dict()
        d["avg_snort_ids_rule_counters"] = self.avg_snort_ids_rule_counters.to_dict()
        d["avg_ossec_ids_alert_counters"] = self.avg_ossec_ids_alert_counters.to_dict()
        return d

    def sort_machines(self) -> None:
        """
        Sorts the machines in the observation

        :return: None
        """
        self.machines = sorted(self.machines, key=lambda x: int(x.ips[0].rsplit(".", 1)[-1]), reverse=False)

    def cleanup(self) -> None:
        """
        Cleans up the machines in the observation

        :return: None
        """
        if self.docker_stats_consumer_thread is not None:
            self.docker_stats_consumer_thread.running = False
            self.docker_stats_consumer_thread.consumer.close()
        if self.client_population_consumer_thread is not None:
            self.client_population_consumer_thread.running = False
            self.client_population_consumer_thread.consumer.close()
        if self.aggregated_snort_ids_log_consumer_thread is not None:
            self.aggregated_snort_ids_log_consumer_thread.running = False
            self.aggregated_snort_ids_log_consumer_thread.consumer.close()
        if self.aggregated_snort_ids_rule_log_consumer_thread is not None:
            self.aggregated_snort_ids_rule_log_consumer_thread.running = False
            self.aggregated_snort_ids_rule_log_consumer_thread.consumer.close()
        if self.aggregated_ossec_ids_log_consumer_thread is not None:
            self.aggregated_ossec_ids_log_consumer_thread.running = False
            self.aggregated_ossec_ids_log_consumer_thread.consumer.close()
        if self.attacker_actions_consumer_thread is not None:
            self.attacker_actions_consumer_thread.running = False
            self.attacker_actions_consumer_thread.consumer.close()
        if self.defender_actions_consumer_thread is not None:
            self.defender_actions_consumer_thread.running = False
            self.defender_actions_consumer_thread.consumer.close()
        if self.aggregated_host_metrics_thread is not None:
            self.aggregated_host_metrics_thread.running = False
        for m in self.machines:
            m.cleanup()

    def reset_metric_lists(self) -> None:
        """
        Resets the metric lists

        :return: None
        """
        if self.aggregated_snort_ids_log_consumer_thread is None or \
                self.aggregated_snort_ids_rule_log_consumer_thread is None or \
                self.aggregated_ossec_ids_log_consumer_thread is None or \
                self.docker_stats_consumer_thread is None or \
                self.client_population_consumer_thread is None or \
                self.aggregated_host_metrics_thread is None:
            raise ValueError("At least one of the objects is None")

        self.aggregated_snort_ids_log_consumer_thread.snort_ids_alert_counters_list = []
        self.aggregated_snort_ids_rule_log_consumer_thread.snort_ids_rule_counters_list = []
        self.aggregated_ossec_ids_log_consumer_thread.ossec_ids_alert_counters_list = []
        self.docker_stats_consumer_thread.docker_stats_list = []
        self.client_population_consumer_thread.client_population_metrics_list = []
        self.aggregated_host_metrics_thread.host_metrics_list = []

    def average_metric_lists(self):
        """
        :return: computes the averages of the metric lists
        """
        if self.aggregated_snort_ids_log_consumer_thread is None or \
                self.aggregated_snort_ids_rule_log_consumer_thread is None or \
                self.aggregated_ossec_ids_log_consumer_thread is None or \
                self.docker_stats_consumer_thread is None or \
                self.client_population_consumer_thread is None or \
                self.aggregated_host_metrics_thread is None:
            raise ValueError("At least one of the objects is None")
        self.avg_snort_ids_alert_counters = \
            self.aggregated_snort_ids_log_consumer_thread.get_aggregated_ids_alert_counters()
        self.avg_snort_ids_rule_counters = \
            self.aggregated_snort_ids_rule_log_consumer_thread.get_aggregated_ids_rule_counters()
        self.avg_ossec_ids_alert_counters = \
            self.aggregated_ossec_ids_log_consumer_thread.get_aggregated_ids_alert_counters()
        self.avg_docker_stats = self.docker_stats_consumer_thread.get_average_docker_stats()
        self.avg_aggregated_host_metrics = self.aggregated_host_metrics_thread.get_average_aggregated_host_metrics()
        self.avg_client_population_metrics = \
            self.client_population_consumer_thread.get_average_client_population_metrics()

    def get_action_ips(self, a: EmulationDefenderAction, emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        Gets the ips of the node that a defender action is targeted for

        :param a: the action
        :param emulation_env_config: the emulation env config
        :return: the ip of the target node
        """
        if a.index == -1:
            return emulation_env_config.topology_config.subnetwork_masks
        if a.index < len(self.machines):
            return self.machines[a.index].ips
        return a.ips

    def copy(self) -> "EmulationDefenderObservationState":
        """
        :return: a copy of the object
        """
        if self.client_population_metrics is None or self.docker_stats is None or \
                self.snort_ids_alert_counters is None or self.ossec_ids_alert_counters is None or \
                self.attacker_actions is None or self.defender_actions is None or \
                self.aggregated_host_metrics is None or \
                self.snort_ids_rule_counters is None or \
                self.avg_aggregated_host_metrics is None:
            raise ValueError("At least of the objects is None")
        c = EmulationDefenderObservationState(
            kafka_config=self.kafka_config,
            client_population_metrics=self.client_population_metrics.copy(), docker_stats=self.docker_stats.copy(),
            snort_ids_alert_counters=self.snort_ids_alert_counters.copy(),
            ossec_ids_alert_counters=self.ossec_ids_alert_counters.copy(),
            attacker_actions=self.attacker_actions.copy(),
            defender_actions=self.defender_actions.copy(), aggregated_host_metrics=self.aggregated_host_metrics.copy(),
            snort_ids_rule_counters=self.snort_ids_rule_counters.copy(),
        )
        c.actions_tried = self.actions_tried.copy()
        c.avg_snort_ids_alert_counters = self.avg_snort_ids_alert_counters.copy()
        c.avg_snort_ids_rule_counters = self.avg_snort_ids_rule_counters.copy()
        c.avg_ossec_ids_alert_counters = self.avg_ossec_ids_alert_counters.copy()
        c.avg_docker_stats = self.avg_docker_stats.copy()
        c.avg_aggregated_host_metrics = self.avg_aggregated_host_metrics.copy()
        c.avg_client_population_metrics = self.avg_client_population_metrics.copy()

        for m in self.machines:
            c.machines.append(m.copy())
        return c

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        if self.attacker_actions is None:
            raise ValueError("attacker_Actions is not iterable")
        if self.defender_actions is None:
            raise ValueError("defender_actions is not iterable")
        return f"client_population_metrics: {self.client_population_metrics}," \
               f"docker_stats: {self.docker_stats}," \
               f"snort_ids_alert_counters: {self.snort_ids_alert_counters}," \
               f"snort_ids_rule_counters: {self.snort_ids_rule_counters}," \
               f"ossec_ids_alert_counters: {self.ossec_ids_alert_counters}," \
               f"aggregated_host_metrics: {self.aggregated_host_metrics}" \
               f"attacker_actions: {list(map(lambda x: str(x), self.attacker_actions))}," \
               f"defender_actions: {list(map(lambda x: str(x), self.defender_actions))}\n," \
               f"avg_snort_ids_alert_counters: {self.avg_snort_ids_alert_counters}," \
               f"avg_snort_ids_rule_counters: {self.avg_snort_ids_rule_counters}," \
               f"avg_ossec_ids_alert_counters: {self.avg_ossec_ids_alert_counters}," \
               f"avg_docker_stats: {self.avg_docker_stats}," \
               f"avg_aggregated_host_metrics: {self.avg_aggregated_host_metrics}," \
               f"avg_client_population_metrics: {self.avg_client_population_metrics}" \
            + "\n".join([str(i) + ":" + str(self.machines[i]) for i in range(len(self.machines))])

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationDefenderObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationDefenderObservationState.from_dict(json.loads(json_str))

    def num_attributes(self) -> int:
        """
        :return: The number of attribute of the DTO
        """
        if self.defender_actions is None:
            raise ValueError("defender_action is None and thus has no size")
        if self.attacker_actions is None:
            raise ValueError("attacker_action is None and thus has no size")
        num_attributes = 0
        if self.client_population_metrics is not None:
            num_attributes = num_attributes + self.client_population_metrics.num_attributes()
        if self.docker_stats is not None:
            num_attributes = num_attributes + self.docker_stats.num_attributes()
        if self.snort_ids_alert_counters is not None:
            num_attributes = num_attributes + self.snort_ids_alert_counters.num_attributes()
        if self.snort_ids_rule_counters is not None:
            num_attributes = num_attributes + 1
        if self.ossec_ids_alert_counters is not None:
            num_attributes = num_attributes + self.ossec_ids_alert_counters.num_attributes()
        if self.aggregated_host_metrics is not None:
            num_attributes = num_attributes + self.aggregated_host_metrics.num_attributes()
        if len(self.defender_actions) > 0:
            num_attributes = num_attributes + len(self.defender_actions) * self.defender_actions[0].num_attributes()
        if len(self.attacker_actions) > 0:
            num_attributes = num_attributes + len(self.attacker_actions) * self.attacker_actions[0].num_attributes()
        return num_attributes

    @staticmethod
    def schema() -> "EmulationDefenderObservationState":
        """
        :return: get the schema of the DTO
        """
        return EmulationDefenderObservationState(kafka_config=KafkaConfig.schema(),
                                                 client_population_metrics=ClientPopulationMetrics.schema(),
                                                 docker_stats=DockerStats.schema(),
                                                 snort_ids_alert_counters=SnortIdsAlertCounters.schema(),
                                                 ossec_ids_alert_counters=OSSECIdsAlertCounters.schema(),
                                                 aggregated_host_metrics=HostMetrics.schema(),
                                                 defender_actions=[EmulationDefenderAction.schema()],
                                                 attacker_actions=[EmulationAttackerAction.schema()],
                                                 snort_ids_rule_counters=SnortIdsRuleCounters.schema())



from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.controllers.traffic_controller import TrafficController
import pytest


class TestTrafficControllerSuite:
    """
    Test Trafic controller
    """

    @pytest.fixture(autouse=True)
    def emulation_env_config_setup(self) -> None:
        """
        Set up emulation environment configuration

        :return: None
        """
        logger = MagicMock()

        # Create mock node traffic configurations
        mock_traffic_config1 = MagicMock()
        mock_traffic_config1.physical_host_ip = "192.168.1.100"
        mock_traffic_config2 = MagicMock()
        mock_traffic_config2.physical_host_ip = "192.168.1.101"
        node_traffic_configs = [mock_traffic_config1, mock_traffic_config2]

        # Create mock traffic config
        traffic_config = MagicMock()
        traffic_config.node_traffic_configs = node_traffic_configs

        # Create mock emulation environment config
        emulation_env_config = MagicMock()
        emulation_env_config.traffic_config = traffic_config
        emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip = "192.168.0.1"
        emulation_env_config.traffic_config.client_population_config.ip = "192.168.8.1"
        emulation_env_config.traffic_config.client_population_config.physical_host_ip = "192.168.1.1"
        emulation_env_config.traffic_config.client_population_config.client_manager_port = 50051
        emulation_env_config.traffic_config.client_population_config.client_manager_log_dir = "/var/log"
        emulation_env_config.traffic_config.client_population_config.client_manager_log_file = "client_manager.log"
        emulation_env_config.traffic_config.client_population_config.client_manager_max_workers = 4
        emulation_env_config.traffic_config.client_population_config.client_time_step_len_seconds = 5
        emulation_env_config.traffic_config.client_population_config.clients = ["client1", "client2"]
        emulation_env_config.traffic_config.client_population_config.workflows_config.workflow_markov_chains = [
            "chain1", "chain2"]
        emulation_env_config.traffic_config.client_population_config.workflows_config.workflow_services = [
            "service1", "service2"]
        emulation_env_config.get_connection.return_value = "mock_connection"
        emulation_env_config.kafka_config.container.get_ips.return_value = ["192.168.0.2"]
        emulation_env_config.kafka_config.kafka_port = 9092
        emulation_env_config.kafka_config.time_step_len_seconds = 5
        self.emulation_env_config = emulation_env_config
        self.logger = logger

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_traffic_manager")
    @patch("time.sleep", return_value=None)
    def test_start_traffic_managers(self, mock_sleep, mock_start_traffic_manager) -> None:
        """
        Test utility method for checking if the traffic manager is running and starting it if it is not running
        on every node

        :param mock_start_traffic_manager: mock_start_traffic_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        TrafficController.start_traffic_managers(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.traffic_config.node_traffic_configs[0].physical_host_ip,
            logger=self.logger)
        mock_start_traffic_manager.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_traffic_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test utility method for starting traffic manager on a specific container

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param mock_sleep: mock_sleep
        :return: None
        """
        self.emulation_env_config.get_connection.return_value = MagicMock()
        node_traffic_config = MagicMock()
        node_traffic_config.docker_gw_bridge_ip = "10.0.0.1"
        node_traffic_config.ip = "10.0.5.1"
        node_traffic_config.traffic_manager_port = 50051
        node_traffic_config.traffic_manager_log_dir = "/var/log/traffic"
        node_traffic_config.traffic_manager_log_file = "traffic.log"
        node_traffic_config.traffic_manager_max_workers = 4
        mock_execute_ssh_cmd.side_effect = [
            (b"", b"", 0),  # Output for stopping old background job
            (b"", b"", 0),  # Output for starting the traffic_manager
        ]
        TrafficController.start_traffic_manager(
            emulation_env_config=self.emulation_env_config, node_traffic_config=node_traffic_config, logger=self.logger
        )
        mock_connect_admin.assert_called()
        stop_cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL
                    + constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.TRAFFIC_MANAGER_FILE_NAME)
        mock_execute_ssh_cmd.assert_any_call(cmd=stop_cmd, conn=self.emulation_env_config.get_connection.return_value)
        start_cmd = constants.COMMANDS.START_TRAFFIC_MANAGER.format(
            node_traffic_config.traffic_manager_port, node_traffic_config.traffic_manager_log_dir,
            node_traffic_config.traffic_manager_log_file, node_traffic_config.traffic_manager_max_workers)
        mock_execute_ssh_cmd.assert_any_call(cmd=start_cmd, conn=self.emulation_env_config.get_connection.return_value)
        self.logger.info.assert_called_with(f"Starting traffic manager on node "
                                            f"{node_traffic_config.docker_gw_bridge_ip} ({node_traffic_config.ip}), "
                                            f"with cmd:{start_cmd}")

    @patch("csle_common.controllers.traffic_controller.TrafficController.stop_traffic_manager")
    def test_stop_traffic_managers(self, mock_stop_traffic_manager) -> None:
        """
        Test utility method for stopping traffic managers on a given server

        :param mock_start_traffic_manager: mock_start_traffic_manager

        :return: None
        """
        TrafficController.stop_traffic_managers(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.traffic_config.node_traffic_configs[0].physical_host_ip,
            logger=self.logger)
        mock_stop_traffic_manager.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    def test_stop_traffic_manager(self, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test utility method for stopping a specific traffic manager

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        self.emulation_env_config.get_connection.return_value = MagicMock()
        node_traffic_config = MagicMock()
        node_traffic_config.docker_gw_bridge_ip = "10.0.0.1"
        node_traffic_config.ip = "10.0.5.1"
        mock_execute_ssh_cmd.return_value = ("output", "error", None)
        TrafficController.stop_traffic_manager(
            emulation_env_config=self.emulation_env_config, node_traffic_config=node_traffic_config, logger=self.logger)
        mock_connect_admin.assert_called_once_with(emulation_env_config=self.emulation_env_config,
                                                   ip=node_traffic_config.docker_gw_bridge_ip)
        self.logger.info.assert_any_call(f"Stopping traffic manager on node "
                                         f"{node_traffic_config.docker_gw_bridge_ip} ({node_traffic_config.ip})")
        self.emulation_env_config.get_connection.assert_called_once_with(ip=node_traffic_config.docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("csle_common.util.emulation_util.EmulationUtil.disconnect_admin")
    @patch("time.sleep", return_value=None)
    def test_start_client_manager(self, mock_sleep, mock_disconnect_admin, mock_execute_ssh_cmd,
                                  mock_connect_admin) -> None:
        """
        Test utility method starting the client manager

        :param mock_disconnect_admin: mock_disconnect_admin
        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_execute_ssh_cmd.side_effect = [
            ("", "", None),  # PS command, client manager not running
            ("", "", None),  # PKILL command
            ("", "", None),  # Start client manager command
        ]
        result = TrafficController.start_client_manager(
            emulation_env_config=self.emulation_env_config, logger=self.logger
        )
        mock_connect_admin.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()
        self.logger.info.assert_called()
        assert result

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("csle_common.util.emulation_util.EmulationUtil.disconnect_admin")
    def test_stop_client_manager(self, mock_disconnect_admin, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test utility method starting the client manager

        :param mock_disconnect_admin: mock_disconnect_admin
        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        mock_execute_ssh_cmd.return_value = ("output", "error", None)
        TrafficController.stop_client_manager(emulation_env_config=self.emulation_env_config, logger=self.logger)
        mock_connect_admin.assert_called_once_with(
            emulation_env_config=self.emulation_env_config,
            ip=self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()
        self.logger.info.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_client_manager")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_clients_dto_by_ip_and_port")
    @patch("grpc.insecure_channel")
    def test_stop_client_population(self, mock_insecure_channel, mock_get_clients_dto,
                                    mock_start_client_manager) -> None:
        """
        Test function for stopping the client arrival process of an emulation

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_get_clients_dto: mock_get_clients_dto
        :param mock_start_client_manager: mock_start_client_manager
        :return: None
        """
        mock_client_dto = MagicMock()
        mock_client_dto.client_process_active = True
        mock_get_clients_dto.return_value = mock_client_dto
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub

        TrafficController.stop_client_population(emulation_env_config=self.emulation_env_config, logger=self.logger)
        self.logger.info.assert_any_call(
            f"Stopping client population on container: "
            f"{self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({self.emulation_env_config.traffic_config.client_population_config.ip})")
        mock_start_client_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config,
                                                          logger=self.logger)
        mock_get_clients_dto.assert_called()
        mock_insecure_channel.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_client_manager")
    @patch("csle_common.controllers.traffic_controller.TrafficController.start_client_population")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_clients_dto_by_ip_and_port")
    @patch("grpc.insecure_channel")
    @patch("time.sleep", return_value=None)
    def test_start_client_producer(self, mock_sleep, mock_insecure_channel, mock_get_clients_dto,
                                   mock_start_client_population, mock_start_client_manager) -> None:
        """
        Test method that starts the Kafka producer for client metrics

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_get_clients_dto: mock_get_clients_dto
        :param mock_start_client_population: mock_start_client_population
        :param mock_start_client_manager: mock_start_client_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_start_client_manager.return_value = True
        mock_client_dto = MagicMock()
        mock_client_dto.producer_active = False
        mock_get_clients_dto.return_value = mock_client_dto
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub

        TrafficController.start_client_producer(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.traffic_config.client_population_config.physical_host_ip,
            logger=self.logger)
        self.logger.info.assert_any_call(
            f"Starting client producer on container:"
            f" {self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({self.emulation_env_config.traffic_config.client_population_config.ip})")
        mock_start_client_manager.assert_called_once_with(emulation_env_config=self.emulation_env_config,
                                                          logger=self.logger)
        mock_start_client_population.assert_called()
        mock_get_clients_dto.assert_called()
        mock_insecure_channel.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_client_manager")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_clients_dto_by_ip_and_port")
    @patch("grpc.insecure_channel")
    def test_stop_client_producer(self, mock_insecure_channel, mock_get_clients_dto, mock_start_client_manager) -> None:
        """
        Test method that stops the Kafka producer for client metrics

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_get_clients_dto: mock_get_clients_dto
        :param mock_start_client_manager: mock_start_client_manager
        :return: None
        """
        mock_start_client_manager.return_value = True
        mock_client_dto = MagicMock()
        mock_client_dto.producer_active = True
        mock_get_clients_dto.return_value = mock_client_dto
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub
        TrafficController.stop_client_producer(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.traffic_config.client_population_config.physical_host_ip,
            logger=self.logger)
        self.logger.info.assert_any_call(
            "Stopping client producer on container: "
            f"{self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({self.emulation_env_config.traffic_config.client_population_config.ip})")
        mock_start_client_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, logger=self.logger)
        mock_get_clients_dto.assert_called()
        mock_insecure_channel.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_client_manager")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_clients_dto_by_ip_and_port")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.client_manager.query_clients.stop_clients")
    @patch("csle_collector.client_manager.query_clients.start_clients")
    @patch("time.sleep", return_value=None)
    def test_start_client_population(self, mock_sleep, mock_start_clients, mock_stop_clients, mock_insecure_channel,
                                     mock_get_clients_dto, mock_start_client_manager) -> None:
        """
        Test method that starts the arrival process of clients

        :param mock_start_clients: mock_start_clients
        :param mock_stop_clients: mock_stop_clients
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_get_clients_dto: mock_get_clients_dto
        :param mock_start_client_manager: mock_start_client_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_start_client_manager.return_value = True
        mock_client_dto = MagicMock()
        mock_client_dto.producer_active = True
        mock_get_clients_dto.return_value = mock_client_dto
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub

        TrafficController.start_client_population(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=self.emulation_env_config.traffic_config.client_population_config.physical_host_ip,
            logger=self.logger)
        self.logger.info.assert_any_call(
            f"Starting client population on container: "
            f"{self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({self.emulation_env_config.traffic_config.client_population_config.ip})")
        mock_start_client_manager.assert_called()
        mock_get_clients_dto.assert_called()
        mock_insecure_channel.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_client_manager")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_clients_dto_by_ip_and_port")
    def test_get_num_active_clients(self, mock_get_clients_dto, mock_start_client_manager) -> None:
        """
        Test method that gets the number of active clients

        :param mock_get_clients_dto: mock_get_clients_dto
        :param mock_start_client_manager: mock_start_client_manager
        :return: None
        """
        mock_client_dto = MagicMock()
        mock_get_clients_dto.return_value = mock_client_dto
        result = TrafficController.get_num_active_clients(
            emulation_env_config=self.emulation_env_config, logger=self.logger)
        mock_start_client_manager.assert_called_once_with(
            emulation_env_config=self.emulation_env_config, logger=self.logger)
        mock_get_clients_dto.assert_called()
        assert result == mock_client_dto

    @patch("grpc.insecure_channel")
    @patch("csle_collector.client_manager.query_clients.get_clients")
    def test_get_clients_dto_by_ip_and_port(self, mock_get_clients, mock_insecure_channel) -> None:
        """
        Test method that sends a request to the ClientManager on a specific container
        to get its status

        :param mock_get_clients: mock_get_clients
        :param mock_insecure_channel: mock_insecure_channel
        :return: None
        """
        ip = "192.168.0.1"
        port = 50051
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub
        mock_status = MagicMock()
        mock_get_clients.return_value = mock_status
        result = TrafficController.get_clients_dto_by_ip_and_port(ip=ip, port=port, logger=self.logger)
        self.logger.info.assert_any_call(
            f"Get client manager status from container with ip: {ip} and client manager port: {port}")
        mock_insecure_channel.assert_called()
        assert result == mock_status

    @patch("csle_common.controllers.traffic_controller.TrafficController.stop_internal_traffic_generator")
    def test_stop_internal_traffic_generators(self, mock_stop_internal_traffic_generator) -> None:
        """
        Test utility function for stopping internal traffic generators

        :param mock_stop_internal_traffic_generator: mock_stop_internal_traffic_generator
        :return: None
        """
        mock_node_traffic_config_1 = MagicMock()
        mock_node_traffic_config_1.physical_host_ip = "192.168.1.1"
        mock_node_traffic_config_2 = MagicMock()
        mock_node_traffic_config_2.physical_host_ip = "192.168.1.2"
        self.emulation_env_config.traffic_config.node_traffic_configs = [mock_node_traffic_config_1,
                                                                         mock_node_traffic_config_2]
        TrafficController.stop_internal_traffic_generators(
            emulation_env_config=self.emulation_env_config, logger=self.logger,
            physical_server_ip=mock_node_traffic_config_1.physical_host_ip)
        mock_stop_internal_traffic_generator.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_traffic_manager")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.traffic_manager.query_traffic_manager.stop_traffic")
    def test_stop_internal_traffic_generator(self, mock_stop_traffic, mock_insecure_channel,
                                             mock_start_traffic_manager) -> None:
        """
        Test utility function for stopping a specific internal traffic generator

        :param mock_stop_traffic: mock_stop_traffic
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_traffic_manager: mock_start_traffic_manager
        :return: None
        """
        mock_node_traffic_config = MagicMock()
        mock_node_traffic_config.docker_gw_bridge_ip = "192.168.0.1"
        mock_node_traffic_config.ip = "192.168.5.1"
        mock_node_traffic_config.traffic_manager_port = 50051
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub
        TrafficController.stop_internal_traffic_generator(
            emulation_env_config=self.emulation_env_config, node_traffic_config=mock_node_traffic_config,
            logger=self.logger)
        mock_start_traffic_manager.assert_called()
        self.logger.info.assert_any_call(f"Stopping traffic generator script, node ip:"
                                         f"{mock_node_traffic_config.docker_gw_bridge_ip} "
                                         f"({mock_node_traffic_config.ip})")
        mock_insecure_channel.assert_called()
        TrafficController.stop_internal_traffic_generator(
            emulation_env_config=self.emulation_env_config,
            node_traffic_config=mock_node_traffic_config, logger=self.logger)
        mock_start_traffic_manager.assert_called()
        self.logger.info.assert_any_call(f"Stopping traffic generator script, node ip:"
                                         f"{mock_node_traffic_config.docker_gw_bridge_ip} "
                                         f"({mock_node_traffic_config.ip})")
        mock_insecure_channel.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_internal_traffic_generator")
    def test_start_internal_traffic_generators(self, mock_start_internal_traffic_generator) -> None:
        """
        Test utility function for starting internal traffic generators

        :param mock_start_internal_traffic_generator: mock_start_internal_traffic_generator
        :return: None
        """
        mock_node_traffic_config_1 = MagicMock()
        mock_node_traffic_config_1.physical_host_ip = "192.168.1.1"
        mock_node_traffic_config_1.ip = "10.0.0.1"
        mock_node_traffic_config_2 = MagicMock()
        mock_node_traffic_config_2.physical_host_ip = "192.168.1.2"
        self.emulation_env_config.traffic_config.node_traffic_configs = [mock_node_traffic_config_1,
                                                                         mock_node_traffic_config_2]
        # Create mock containers
        mock_container_1 = MagicMock()
        mock_container_1.get_ips.return_value = ["10.0.0.1"]
        mock_container_2 = MagicMock()
        mock_container_2.get_ips.return_value = ["10.0.0.2"]
        self.emulation_env_config.containers_config.containers = [mock_container_1, mock_container_2]
        TrafficController.start_internal_traffic_generators(
            emulation_env_config=self.emulation_env_config,
            physical_server_ip=mock_node_traffic_config_1.physical_host_ip, logger=self.logger)
        mock_start_internal_traffic_generator.assert_called()

    @patch("csle_common.controllers.traffic_controller.TrafficController.start_traffic_manager")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.traffic_manager.query_traffic_manager.start_traffic")
    def test_start_internal_traffic_generator(self, mock_start_traffic, mock_insecure_channel,
                                              mock_start_traffic_manager) -> None:
        """
        Test utility function for starting internal traffic generators

        :param mock_start_traffic: mock_start_traffic
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_traffic_manager: mock_start_traffic_manager
        :return: None
        """
        mock_node_traffic_config = MagicMock()
        mock_node_traffic_config.docker_gw_bridge_ip = "192.168.0.1"
        mock_node_traffic_config.traffic_manager_port = 50051

        mock_container = MagicMock()
        mock_container.ips_and_networks = [("10.0.0.1", MagicMock(subnet_mask="255.255.255.0"))]
        mock_container.get_ips.return_value = ["10.0.0.1"]

        mock_container_1 = MagicMock()
        mock_container_1.ips_and_networks = [("10.0.0.2", MagicMock(subnet_mask="255.255.255.0"))]
        mock_container_1.get_ips.return_value = ["10.0.0.2"]

        mock_container_2 = MagicMock()
        mock_container_2.ips_and_networks = [("10.0.0.3", MagicMock(subnet_mask="255.255.255.0"))]
        mock_container_2.get_ips.return_value = ["10.0.0.3"]

        self.emulation_env_config.containers_config.containers = [mock_container_1, mock_container_2]

        mock_node_traffic_config_1 = MagicMock()
        mock_node_traffic_config_1.ip = "10.0.0.2"
        mock_node_traffic_config_1.commands = ["cmd1"]

        mock_node_traffic_config_2 = MagicMock()
        mock_node_traffic_config_2.ip = "10.0.0.3"
        mock_node_traffic_config_2.commands = ["cmd2"]

        self.emulation_env_config.traffic_config.node_traffic_configs = [
            mock_node_traffic_config_1, mock_node_traffic_config_2]
        self.emulation_env_config.traffic_config.client_population_config.client_time_step_len_seconds = 5

        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub

        TrafficController.start_internal_traffic_generator(
            emulation_env_config=self.emulation_env_config,
            node_traffic_config=mock_node_traffic_config, container=mock_container, logger=self.logger)
        mock_start_traffic_manager.assert_called()
        self.logger.info.assert_any_call(f"Starting traffic generator script, node ip:"
                                         f"{mock_node_traffic_config.docker_gw_bridge_ip} "
                                         f"({mock_node_traffic_config.ip})")
        mock_insecure_channel.assert_called()

    def test_get_client_managers_ips(self) -> None:
        """
        Test method that extracts the IPS of the Client managers in a given emulation

        :return: None
        """
        result = TrafficController.get_client_managers_ips(emulation_env_config=self.emulation_env_config)
        assert result == [self.emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip]

    def test_get_client_managers_ports(self) -> None:
        """
        Test method that extracts the ports of the Client managers in a given emulation

        :return: None
        """
        result = TrafficController.get_client_managers_ports(emulation_env_config=self.emulation_env_config)
        assert result == [50051]

    @patch("csle_common.controllers.traffic_controller.TrafficController.get_client_managers_ips")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_client_managers_ports")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_clients_dto_by_ip_and_port")
    @patch("csle_collector.client_manager.client_manager_util.ClientManagerUtil.clients_dto_empty")
    def test_get_client_managers_info(self, mock_clients_dto_empty, mock_get_clients_dto_by_ip_and_port,
                                      mock_get_client_managers_ports, mock_get_client_managers_ips) -> None:
        """
        Test method that extracts the information of the Client managers for a given emulation

        :param mock_clients_dto_empty: mock_clients_dto_empty
        :param mock_get_clients_dto_by_ip_and_port: mock_get_clients_dto_by_ip_and_port
        :param mock_get_client_managers_ports: mock_get_client_managers_ports
        :param mock_get_client_managers_ips: mock_get_client_managers_ips
        :return: None
        """
        self.emulation_env_config.execution_id = "test_execution_id"
        self.emulation_env_config.name = "test_emulation"
        self.emulation_env_config.traffic_config.client_population_config.client_manager_port = 50051

        active_ips = ["192.168.0.1", "192.168.0.2"]
        ports = [50051, 50051]
        mock_get_client_managers_ips.return_value = active_ips
        mock_get_client_managers_ports.return_value = ports

        mock_status = MagicMock()
        mock_get_clients_dto_by_ip_and_port.side_effect = [mock_status, Exception("Test Exception")]
        mock_clients_dto_empty.return_value = "empty_dto"

        result = TrafficController.get_client_managers_info(
            emulation_env_config=self.emulation_env_config, active_ips=active_ips, logger=self.logger)

        mock_get_client_managers_ips.assert_called_once_with(emulation_env_config=self.emulation_env_config)
        mock_get_client_managers_ports.assert_called_once_with(emulation_env_config=self.emulation_env_config)
        mock_get_clients_dto_by_ip_and_port.assert_any_call(ip=active_ips[0], port=ports[0], logger=self.logger)
        mock_get_clients_dto_by_ip_and_port.assert_any_call(ip=active_ips[1], port=ports[1], logger=self.logger)
        assert result

    @patch("grpc.insecure_channel")
    @patch("csle_collector.traffic_manager.query_traffic_manager.get_traffic_status")
    def test_get_traffic_manager_status_by_port_and_ip(self, mock_get_traffic_status, mock_insecure_channel) -> None:
        """
        Test method that sends a request to the TrafficManager on a specific container
        to get its status

        :param mock_get_traffic_status: mock_get_traffic_status
        :param mock_insecure_channel: mock_insecure_channel
        :return: None
        """
        ip = "192.168.0.1"
        port = 50051
        mock_channel = MagicMock()
        mock_stub = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = mock_channel
        mock_channel.return_value = mock_stub
        mock_status = MagicMock()
        mock_get_traffic_status.return_value = mock_status
        result = TrafficController.get_traffic_manager_status_by_port_and_ip(ip=ip, port=port)
        mock_insecure_channel.assert_called()
        assert result == mock_status

    def test_get_traffic_managers_ips(self) -> None:
        """
        Test method that extracts the ips of the traffic managers in a given emulation

        :return: None
        """
        mock_node_traffic_config_1 = MagicMock()
        mock_node_traffic_config_1.docker_gw_bridge_ip = "192.168.0.1"
        mock_node_traffic_config_2 = MagicMock()
        mock_node_traffic_config_2.docker_gw_bridge_ip = "192.168.0.2"
        self.emulation_env_config.traffic_config.node_traffic_configs = [
            mock_node_traffic_config_1, mock_node_traffic_config_2]
        result = TrafficController.get_traffic_managers_ips(emulation_env_config=self.emulation_env_config)
        assert result == [mock_node_traffic_config_1.docker_gw_bridge_ip,
                          mock_node_traffic_config_2.docker_gw_bridge_ip]

    def test_get_traffic_managers_ports(self) -> None:
        """
        Test method that extracts the ports of the Traffic managers in a given emulation

        :return: None
        """
        mock_node_traffic_config_1 = MagicMock()
        mock_node_traffic_config_1.traffic_manager_port = 50051
        mock_node_traffic_config_2 = MagicMock()
        mock_node_traffic_config_2.traffic_manager_port = 50052
        self.emulation_env_config.traffic_config.node_traffic_configs = [
            mock_node_traffic_config_1, mock_node_traffic_config_2]
        result = TrafficController.get_traffic_managers_ports(emulation_env_config=self.emulation_env_config)
        assert result == [mock_node_traffic_config_1.traffic_manager_port,
                          mock_node_traffic_config_2.traffic_manager_port]

    @patch("csle_common.controllers.traffic_controller.TrafficController.get_traffic_managers_ips")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_traffic_managers_ports")
    @patch("csle_common.controllers.traffic_controller.TrafficController.get_traffic_manager_status_by_port_and_ip")
    @patch("csle_collector.traffic_manager.traffic_manager_util.TrafficManagerUtil.traffic_dto_empty")
    @patch("csle_common.util.emulation_util.EmulationUtil.physical_ip_match")
    @patch("time.sleep", return_value=None)
    def test_get_traffic_managers_info(
            self, mock_sleep, mock_physical_ip_match, mock_traffic_dto_empty,
            mock_get_traffic_manager_status_by_port_and_ip, mock_get_traffic_managers_ports,
            mock_get_traffic_managers_ips) -> None:
        """
        Test method that extracts the information of the traffic managers for a given emulation

        :param mock_physical_ip_match: mock_physical_ip_match
        :param mock_traffic_dto_empty: mock_traffic_dto_empty
        :param mock_get_traffic_manager_status_by_port_and_ip: mock_get_traffic_manager_status_by_port_and_ip
        :param mock_get_traffic_managers_ports: mock_get_traffic_managers_ports
        :param mock_get_traffic_managers_ips: mock_get_traffic_managers_ips
        :param mock_sleep: mock_sleep
        :return: None
        """
        self.emulation_env_config.execution_id = "test_execution_id"
        self.emulation_env_config.name = "test_emulation"

        mock_node_traffic_config_1 = MagicMock()
        mock_node_traffic_config_1.docker_gw_bridge_ip = "192.168.0.1"
        mock_node_traffic_config_1.traffic_manager_port = 50051
        mock_node_traffic_config_1.ip = "10.0.0.1"

        mock_node_traffic_config_2 = MagicMock()
        mock_node_traffic_config_2.docker_gw_bridge_ip = "192.168.0.2"
        mock_node_traffic_config_2.traffic_manager_port = 50052
        mock_node_traffic_config_2.ip = "10.0.0.2"

        self.emulation_env_config.traffic_config.node_traffic_configs = [
            mock_node_traffic_config_1, mock_node_traffic_config_2]

        mock_get_traffic_managers_ips.return_value = [mock_node_traffic_config_1.docker_gw_bridge_ip,
                                                      mock_node_traffic_config_2.docker_gw_bridge_ip]
        mock_get_traffic_managers_ports.return_value = [mock_node_traffic_config_1.traffic_manager_port,
                                                        mock_node_traffic_config_2.traffic_manager_port]

        mock_status = MagicMock()
        mock_get_traffic_manager_status_by_port_and_ip.side_effect = [mock_status, Exception("Test Exception")]
        mock_traffic_dto_empty.return_value = "empty_dto"
        mock_physical_ip_match.return_value = True

        active_ips = [mock_node_traffic_config_1.docker_gw_bridge_ip, mock_node_traffic_config_2.docker_gw_bridge_ip]
        physical_host_ip = mock_node_traffic_config_1.ip

        result = TrafficController.get_traffic_managers_info(
            emulation_env_config=self.emulation_env_config,
            active_ips=active_ips, physical_host_ip=physical_host_ip, logger=self.logger)
        mock_get_traffic_managers_ips.assert_called_once_with(emulation_env_config=self.emulation_env_config)
        mock_get_traffic_managers_ports.assert_called_once_with(emulation_env_config=self.emulation_env_config)
        mock_get_traffic_manager_status_by_port_and_ip.assert_any_call(
            port=mock_node_traffic_config_1.traffic_manager_port,
            ip=mock_node_traffic_config_1.docker_gw_bridge_ip)
        mock_get_traffic_manager_status_by_port_and_ip.assert_any_call(
            port=mock_node_traffic_config_2.traffic_manager_port, ip=mock_node_traffic_config_2.docker_gw_bridge_ip)
        assert result



from typing import List, Dict, Any
from csle_base.json_serializable import JSONSerializable


class EmpiricalConditional(JSONSerializable):
    """
    A DTO representing an empirical conditional distribution
    """

    def __init__(self, conditional_name: str, metric_name: str,
                 sample_space: List[int],
                 probabilities: List[float]) -> None:
        """
        Initializes the DTO

        :param conditional_name: the name of the conditional
        :param metric_name: the name of the metric
        :param sample_space: the sample space (the domain of the distribution)
        :param probabilities: the probability distribution
        """
        self.conditional_name = conditional_name
        self.probabilities = probabilities
        assert round(sum(probabilities), 2) == 1
        self.metric_name = metric_name
        self.sample_space = sample_space

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmpiricalConditional":
        """
        Converts a dict representation of the DTO into an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        return EmpiricalConditional(
            conditional_name=d["conditional_name"], metric_name=d["metric_name"],
            sample_space=d["sample_space"], probabilities=d["probabilities"]
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the DTO
        """
        d: Dict[str, Any] = {}
        d["conditional_name"] = self.conditional_name
        d["metric_name"] = self.metric_name
        d["sample_space"] = self.sample_space
        d["probabilities"] = self.probabilities
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return f"conditional_name:{self.conditional_name}, metric_name: {self.metric_name}, " \
               f"sample_space: {self.sample_space}, probabilities: {self.probabilities}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmpiricalConditional":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmpiricalConditional.from_dict(json.loads(json_str))



from typing import List, Dict, Any
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.node_resources_config import NodeResourcesConfig
from csle_common.dao.emulation_config.node_firewall_config import NodeFirewallConfig
from csle_common.dao.emulation_config.kafka_topic import KafkaTopic
from csle_base.json_serializable import JSONSerializable


class KafkaConfig(JSONSerializable):
    """
    Represents the configuration of the Kafka node in a CSLE emulation
    """

    def __init__(self, container: NodeContainerConfig, resources: NodeResourcesConfig,
                 firewall_config: NodeFirewallConfig,
                 topics: List[KafkaTopic],
                 kafka_manager_log_file: str, kafka_manager_log_dir: str, kafka_manager_max_workers: int,
                 kafka_port: int = 9092, kafka_port_external: int = 9292,
                 time_step_len_seconds: int = 15, kafka_manager_port: int = 50051,
                 version: str = "0.0.1") -> None:
        """
        Initializes the DTO

        :param container: the container for the Kafka server
        :param network: the network
        :param kafka_port: the port that the Kafka server is listening to
        :param kafka_port_external: the external port that the Kafka server is listening to
        :param kafka_manager_port: the default port for gRPC
        :param time_step_len_seconds: the length of a time-step (period for logging)
        :param firewall_config: the firewall configuration
        :param container: the container
        :param topics: list of kafka topics
        :param version: the version
        :param kafka_manager_log_file: log file of the kafka manager
        :param kafka_manager_log_dir: log dir of the kafka manager
        :param kafka_manager_max_workers: maximum number of GRPC workers of the kafka manager
        """
        self.kafka_port = kafka_port
        self.kafka_manager_port = kafka_manager_port
        self.time_step_len_seconds = time_step_len_seconds
        self.version = version
        self.container = container
        self.resources = resources
        self.kafka_port_external = kafka_port_external
        self.topics = topics
        self.firewall_config = firewall_config
        self.kafka_manager_log_file = kafka_manager_log_file
        self.kafka_manager_log_dir = kafka_manager_log_dir
        self.kafka_manager_max_workers = kafka_manager_max_workers

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "KafkaConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = KafkaConfig(
            container=NodeContainerConfig.from_dict(d["container"]),
            resources=NodeResourcesConfig.from_dict(d["resources"]),
            topics=list(map(lambda x: KafkaTopic.from_dict(x), d["topics"])),
            kafka_port=d["kafka_port"], time_step_len_seconds=d["time_step_len_seconds"],
            kafka_manager_port=d["kafka_manager_port"],
            version=d["version"],
            firewall_config=NodeFirewallConfig.from_dict(d["firewall_config"]),
            kafka_manager_log_file=d["kafka_manager_log_file"],
            kafka_manager_log_dir=d["kafka_manager_log_dir"],
            kafka_manager_max_workers=d["kafka_manager_max_workers"],
            kafka_port_external=d["kafka_port_external"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["container"] = self.container.to_dict()
        d["resources"] = self.resources.to_dict()
        d["kafka_port"] = self.kafka_port
        d["kafka_manager_port"] = self.kafka_manager_port
        d["time_step_len_seconds"] = self.time_step_len_seconds
        d["version"] = self.version
        d["topics"] = list(map(lambda x: x.to_dict(), self.topics))
        d["firewall_config"] = self.firewall_config.to_dict()
        d["kafka_manager_max_workers"] = self.kafka_manager_max_workers
        d["kafka_manager_log_dir"] = self.kafka_manager_log_dir
        d["kafka_manager_log_file"] = self.kafka_manager_log_file
        d["kafka_port_external"] = self.kafka_port_external
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"container: {self.container}, " \
               f"kafka server port :{self.kafka_port}, version: {self.version}, resources: {self.resources}, " \
               f"topics: {','.join(list(map(lambda x: str(x), self.topics)))}, " \
               f"kafka_manager_port:{self.kafka_manager_port}, time_step_len_seconds: {self.time_step_len_seconds}, " \
               f"firewall_config: {self.firewall_config}, " \
               f"kafka_manager_log_file: {self.kafka_manager_log_file}, " \
               f"kafka_manager_log_dir: {self.kafka_manager_log_dir}, " \
               f"kafka_manager_max_workers: {self.kafka_manager_max_workers}," \
               f"kafka_port_external: {self.kafka_port_external}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "KafkaConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return KafkaConfig.from_dict(json.loads(json_str))

    def copy(self) -> "KafkaConfig":
        """
        :return: a copy of the DTO
        """
        return KafkaConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "KafkaConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :param physical_servers: the physical servers of the execution
        :return: the new config
        """
        config = self.copy()
        config.container = config.container.create_execution_config(ip_first_octet=ip_first_octet,
                                                                    physical_servers=physical_servers)
        config.resources = config.resources.create_execution_config(ip_first_octet=ip_first_octet)
        config.firewall_config = config.firewall_config.create_execution_config(ip_first_octet=ip_first_octet)
        return config

    @staticmethod
    def schema() -> "KafkaConfig":
        """
        :return: get the schema of the DTO
        """
        return KafkaConfig(container=NodeContainerConfig.schema(), resources=NodeResourcesConfig.schema(),
                           firewall_config=NodeFirewallConfig.schema(), topics=[KafkaTopic.schema()],
                           kafka_manager_max_workers=10, kafka_manager_log_dir="/",
                           kafka_manager_log_file="kafka_manager.log")



from typing import Dict, Any
from csle_base.json_serializable import JSONSerializable


class PlayerConfig(JSONSerializable):
    """
    DTO representing the configuration of a player in a simulation environment in CSLE
    """

    def __init__(self, name: str, id: int, descr: str = ""):
        """
        Initializes the DTO

        :param name: the name of the player
        :param id: the id of the player
        :param descr: the description of the player
        """
        self.name = name
        self.id = id
        self.descr = descr

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "PlayerConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = PlayerConfig(name=d["name"], id=int(d["id"]), descr=d["descr"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["id"] = self.id
        d["descr"] = self.descr
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name: {self.name}, id: {self.id}, descr: {self.descr}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "PlayerConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return PlayerConfig.from_dict(json.loads(json_str))



from typing import Tuple, List, Any
import time
import paramiko
import telnetlib
from ftplib import FTP
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_observation.attacker.emulation_attacker_machine_observation_state \
    import EmulationAttackerMachineObservationState
from csle_common.tunneling.forward_tunnel_thread import ForwardTunnelThread
from csle_common.dao.emulation_config.credential import Credential
from csle_common.util.emulation_util import EmulationUtil
from csle_common.dao.emulation_config.connection_setup_dto import ConnectionSetupDTO
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.logging.log import Logger


class ConnectionUtil:
    """
    Class containing utility functions for the connection-related functionality to the emulation
    """

    @staticmethod
    def login_service_helper(s: EmulationEnvState, a: EmulationAttackerAction, alive_check,
                             service_name: str) -> Tuple[EmulationEnvState, bool]:
        """
        Helper function for logging in to a network service in the emulation

        :param s: the current state
        :param a: the action of the login
        :param alive_check:  the function to check whether current connections are alive or not
        :param service_name: name of the service to login to
        :return: s_prime, new connection (bool)
        """
        total_cost = 0.0
        target_machine = None
        non_used_credentials: List[Credential] = []
        root = False
        non_failed_credentials: List[Credential] = []
        if s.attacker_obs_state is None:
            raise ValueError("Could not find any EmulationAttackerObservationState object")
        for m in s.attacker_obs_state.machines:
            if m.ips == a.ips:
                target_machine = m.copy()
                break
        # Check if already logged in
        if target_machine is not None:
            alive_connections = []
            root = False
            connected = False
            connections: List[EmulationConnectionObservationState] = []
            if service_name == constants.TELNET.SERVICE_NAME:
                connections = target_machine.telnet_connections
            elif service_name == constants.SSH.SERVICE_NAME:
                connections = target_machine.ssh_connections
            elif service_name == constants.FTP.SERVICE_NAME:
                connections = target_machine.ftp_connections
            for c in connections:
                if alive_check(c.conn):
                    connected = True
                    alive_connections.append(c)
                    if c.root:
                        root = c.root
                else:
                    if c.tunnel_thread is not None:
                        # stop the tunnel thread that does port forwarding
                        c.tunnel_thread.forward_server.shutdown()
            if len(target_machine.logged_in_services) == 1 and target_machine.logged_in_services[0] == service_name:
                target_machine.logged_in = connected
            if len(target_machine.root_services) == 1 and target_machine.root_services[0] == service_name:
                target_machine.root = root
            if not connected and service_name in target_machine.logged_in_services:
                target_machine.logged_in_services.remove(service_name)
            if not root and service_name in target_machine.root_services:
                target_machine.root_services.remove(service_name)
            non_used_credentials = []
            root = False
            for cr in target_machine.shell_access_credentials:
                if cr.service == service_name:
                    already_logged_in = False
                    for c in alive_connections:
                        if not root and c.root:
                            root = True
                        if c.credential.username == cr.username:
                            already_logged_in = True
                    if not already_logged_in:
                        non_used_credentials.append(cr)

            if service_name == constants.TELNET.SERVICE_NAME:
                target_machine.telnet_connections = alive_connections
            elif service_name == constants.SSH.SERVICE_NAME:
                target_machine.ssh_connections = alive_connections
            elif service_name == constants.FTP.SERVICE_NAME:
                target_machine.ftp_connections = alive_connections

        # Check cached connections
        non_used_nor_cached_credentials = []
        if target_machine is None:
            raise ValueError("Target machine is None")
        for cr in non_used_credentials:
            if cr.service == constants.SSH.SERVICE_NAME:
                for ip in target_machine.ips:
                    key = (ip, cr.username, cr.port)
                    if key in s.attacker_cached_ssh_connections:
                        c = s.attacker_cached_ssh_connections[key]
                        target_machine.ssh_connections.append(c)
                        target_machine.logged_in = True
                        target_machine.root = c.root
                        if cr.service not in target_machine.logged_in_services:
                            target_machine.logged_in_services.append(cr.service)
                        if cr.service not in target_machine.root_services:
                            target_machine.root_services.append(cr.service)
                        break
                    else:
                        non_used_nor_cached_credentials.append(cr)
            elif cr.service == constants.TELNET.SERVICE_NAME:
                for ip in target_machine.ips:
                    key = (ip, cr.username, cr.port)
                    if key in s.attacker_cached_telnet_connections:
                        c = s.attacker_cached_telnet_connections[key]
                        target_machine.telnet_connections.append(c)
                        target_machine.logged_in = True
                        target_machine.root = c.root
                        if cr.service not in target_machine.logged_in_services:
                            target_machine.logged_in_services.append(cr.service)
                        if cr.service not in target_machine.root_services:
                            target_machine.root_services.append(cr.service)
                        break
                    else:
                        non_used_nor_cached_credentials.append(cr)
            elif cr.service == constants.FTP.SERVICE_NAME:
                for ip in target_machine.ips:
                    key = (ip, cr.username, cr.port)
                    if key in s.attacker_cached_ftp_connections:
                        c = s.attacker_cached_ftp_connections[key]
                        target_machine.ftp_connections.append(c)
                        target_machine.logged_in = True
                        target_machine.root = c.root
                        if cr.service not in target_machine.logged_in_services:
                            target_machine.logged_in_services.append(cr.service)
                        if cr.service not in target_machine.root_services:
                            target_machine.root_services.append(cr.service)
                        break
                    else:
                        non_used_nor_cached_credentials.append(cr)

        if target_machine is None or root or len(non_used_nor_cached_credentials) == 0:
            s_prime = s
            if target_machine is not None:
                if s.attacker_obs_state is None or s_prime.attacker_obs_state is None:
                    raise ValueError("EmulationAttackerObservationState is None.")
                attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
                    s.attacker_obs_state.machines, [target_machine],
                    emulation_env_config=s.emulation_env_config, action=a)
                s_prime.attacker_obs_state.machines = attacker_machine_observations

            return s_prime, False

        # If not logged in and there are credentials, setup a new connection
        agent_cr = Credential(
            username=constants.AGENT.USER, pw=constants.AGENT.PW, root=True,
            protocol=TransportProtocol.TCP, service=constants.SSH.SERVICE_NAME, port=constants.SSH.DEFAULT_PORT)
        proxy_connections = [EmulationConnectionObservationState(
            conn=s.emulation_env_config.get_hacker_connection(), credential=agent_cr,
            root=True, port=22, service=constants.SSH.SERVICE_NAME, proxy=None,
            ip=s.emulation_env_config.containers_config.agent_ip)]
        for m in s.attacker_obs_state.machines:
            ssh_connections_sorted_by_root = sorted(
                m.ssh_connections,
                key=lambda x: (constants.SSH_BACKDOOR.BACKDOOR_PREFIX in x.credential.username, x.root,
                               x.credential.username),
                reverse=True)
            if len(ssh_connections_sorted_by_root) > 0:
                proxy_connections.append(ssh_connections_sorted_by_root[0])

        if service_name == constants.SSH.SERVICE_NAME:
            connection_setup_dto = ConnectionUtil._ssh_setup_connection(
                a=a,
                credentials=non_used_nor_cached_credentials,
                proxy_connections=proxy_connections, s=s)
        elif service_name == constants.TELNET.SERVICE_NAME:
            connection_setup_dto = ConnectionUtil._telnet_setup_connection(
                a=a,
                credentials=non_used_nor_cached_credentials,
                proxy_connections=proxy_connections, s=s)
        elif service_name == constants.FTP.SERVICE_NAME:
            connection_setup_dto = ConnectionUtil._ftp_setup_connection(
                a=a,
                credentials=non_used_nor_cached_credentials,
                proxy_connections=proxy_connections, s=s)
        else:
            raise ValueError(f"Service: {service_name} not recognized")

        s_prime = s
        if len(connection_setup_dto.non_failed_credentials) > 0:
            total_cost += connection_setup_dto.total_time
            if connection_setup_dto.connected:
                root = False
                target_machine.logged_in = True
                target_machine.shell_access_credentials = non_failed_credentials
                if service_name not in target_machine.logged_in_services:
                    target_machine.logged_in_services.append(service_name)
                if service_name not in target_machine.root_services:
                    target_machine.root_services.append(service_name)
                for i in range(len(connection_setup_dto.target_connections)):
                    # Check if root
                    c_root = False
                    cost = 0.0
                    if service_name == constants.SSH.SERVICE_NAME:
                        c_root, cost = ConnectionUtil._ssh_finalize_connection(
                            target_machine=target_machine, i=i,
                            connection_setup_dto=connection_setup_dto)
                    elif service_name == constants.TELNET.SERVICE_NAME:
                        c_root, cost = ConnectionUtil._telnet_finalize_connection(
                            target_machine=target_machine, i=i,
                            connection_setup_dto=connection_setup_dto)
                    elif service_name == constants.FTP.SERVICE_NAME:
                        c_root, cost = ConnectionUtil._ftp_finalize_connection(
                            target_machine=target_machine, i=i, connection_setup_dto=connection_setup_dto)
                    total_cost += cost
                    if c_root:
                        root = True
            target_machine.root = root
            attacker_machine_observations = EnvDynamicsUtil.merge_new_obs_with_old(
                s.attacker_obs_state.machines, [target_machine],
                emulation_env_config=s.emulation_env_config, action=a)
            if s_prime.attacker_obs_state is None:
                raise ValueError("Could not find EmulationAttackerObservationState object")
            s_prime.attacker_obs_state.machines = attacker_machine_observations
        else:
            target_machine.shell_access = False
            target_machine.shell_access_credentials = []

        return s_prime, False

    @staticmethod
    def _ssh_setup_connection(a: EmulationAttackerAction,
                              credentials: List[Credential],
                              proxy_connections: List[EmulationConnectionObservationState],
                              s: EmulationEnvState) -> ConnectionSetupDTO:
        """
        Helper function for setting up a SSH connection

        :param a: the action of the connection
        :param credentials: list of credentials to try
        :param proxy_connections: list of proxy connections to try
        :param s: env state
        :return: a DTO with connection setup information
        """
        start = time.time()
        target_connections: List[Any] = []
        non_failed_credentials: List[Credential] = []
        proxies: List[EmulationConnectionObservationState] = []
        ports: List[int] = []
        forward_ports: List[int] = []
        interactive_shells: List[Any] = []
        tunnel_threads: List[ForwardTunnelThread] = []
        connected: bool = False
        ip: str = ""
        for proxy_conn in proxy_connections:
            if proxy_conn.ip is None:
                raise ValueError("EmulationConnectionObservationState is None")
            if proxy_conn.ip != s.emulation_env_config.containers_config.agent_ip:
                m = s.get_attacker_machine(proxy_conn.ip)
                if m is None or not a.ips_match(list(m.reachable)) or m.ips_match(a.ips):
                    continue
            else:
                if s.attacker_obs_state is None or s.attacker_obs_state.agent_reachable is None:
                    raise ValueError("EmulationAttackerObservationState is None")
                if not a.ips_match(list(s.attacker_obs_state.agent_reachable)):
                    continue
            for cr in credentials:
                for ip in a.ips:
                    if cr.service == constants.SSH.SERVICE_NAME:
                        try:
                            agent_addr = (proxy_conn.ip, cr.port)
                            target_addr = (ip, cr.port)
                            agent_transport = proxy_conn.conn.get_transport()
                            relay_channel = agent_transport.open_channel(constants.SSH.DIRECT_CHANNEL, target_addr,
                                                                         agent_addr)
                            target_conn = paramiko.SSHClient()
                            target_conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                            if target_conn is None:
                                raise ValueError("Connection failed")
                            target_conn.connect(ip, username=cr.username, password=cr.pw, sock=relay_channel)
                            transport = target_conn.get_transport()
                            if transport is not None:
                                transport.set_keepalive(5)
                            else:
                                raise ValueError("Connection failed")
                            connected = True
                            credentials.append(cr)
                            target_connections.append(target_conn)
                            proxies.append(proxy_conn)
                            non_failed_credentials.append(cr)
                            ip = ip
                            ports.append(cr.port)
                            break
                        except Exception as e:
                            Logger.__call__().get_logger().warning("SSH exception :{}".format(str(e)))
                            Logger.__call__().get_logger().warning("user:{}, pw:{}".format(cr.username, cr.pw))
                            if s.attacker_obs_state is None:
                                raise ValueError("EmulationAttackerObservationState is None")
                            Logger.__call__().get_logger().warning("Agent reachable:{}".format(
                                s.attacker_obs_state.agent_reachable))
                            Logger.__call__().get_logger().warning("Action:{}, {}, {}".format(a.id, a.ips, a.descr))
                    else:
                        non_failed_credentials.append(cr)
                if connected:
                    break
            if connected:
                break
        end = time.time()
        connection_setup_dto = ConnectionSetupDTO(
            connected=connected, credentials=credentials, target_connections=target_connections,
            proxies=proxies, non_failed_credentials=non_failed_credentials, ip=ip,
            ports=ports, forward_ports=forward_ports, interactive_shells=interactive_shells,
            tunnel_threads=tunnel_threads, total_time=end - start)
        return connection_setup_dto

    @staticmethod
    def _ssh_finalize_connection(target_machine: EmulationAttackerMachineObservationState,
                                 connection_setup_dto: ConnectionSetupDTO, i: int) -> Tuple[bool, float]:
        """
        Helper function for finalizing a SSH connection and setting up the DTO

        :param target_machine: the target machine to connect to
        :param i: current index
        :param connection_setup_dto: DTO with connection setup information
        :return: boolean whether the connection has root privileges or not, cost
        """
        start = time.time()
        root = False
        for j in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT):
            outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(
                cmd="sudo -l", conn=connection_setup_dto.target_connections[i])
            root = False
            if f"{connection_setup_dto.credentials[i].username} may not run sudo" not in errdata.decode("utf-8") \
                    and ("(ALL) NOPASSWD: ALL" in outdata.decode("utf-8") or
                         "(ALL : ALL) ALL" in outdata.decode("utf-8")):
                root = True
                target_machine.root = True
                break
            else:
                time.sleep(constants.ENV_CONSTANTS.SLEEP_RETRY)
        connection_dto = EmulationConnectionObservationState(conn=connection_setup_dto.target_connections[i],
                                                             credential=connection_setup_dto.credentials[i],
                                                             root=root,
                                                             service=constants.SSH.SERVICE_NAME,
                                                             port=connection_setup_dto.ports[i],
                                                             proxy=connection_setup_dto.proxies[i],
                                                             ip=connection_setup_dto.ip)
        target_machine.ssh_connections.append(connection_dto)
        end = time.time()
        total_time = end - start
        return root, total_time

    @staticmethod
    def _telnet_setup_connection(a: EmulationAttackerAction,
                                 credentials: List[Credential],
                                 proxy_connections: List[EmulationConnectionObservationState],
                                 s: EmulationEnvState) -> ConnectionSetupDTO:
        """
        Helper function for setting up a Telnet connection to a target machine

        :param a: the action of the connection
        :param credentials: list of credentials to try
        :param proxy_connections: proxy connections
        :param s: env state
        :return: a DTO with the connection setup information
        """
        target_connections: List[Any] = []
        non_failed_credentials: List[Credential] = []
        proxies: List[EmulationConnectionObservationState] = []
        ports: List[int] = []
        forward_ports: List[int] = []
        interactive_shells: List[Any] = []
        tunnel_threads: List[ForwardTunnelThread] = []
        connected: bool = False
        ip: str = ""
        start = time.time()
        for proxy_conn in proxy_connections:
            if proxy_conn.ip != s.emulation_env_config.containers_config.agent_ip:
                if proxy_conn.ip is None:
                    raise ValueError("EmulationConnectionObservationState is None")
                m = s.get_attacker_machine(proxy_conn.ip)
                if m is None:
                    continue
                if m.ips == a.ips:
                    continue
                no_match = True
                for ip in a.ips:
                    if ip in m.reachable:
                        no_match = False
                if no_match:
                    continue
            else:
                if s.attacker_obs_state is None or s.attacker_obs_state.agent_reachable is None:
                    raise ValueError("EmulationAttackerObservationState or agent_reachable is None")
                if not a.ips_match(list(s.attacker_obs_state.agent_reachable)):
                    continue
            for cr in credentials:
                for ip in a.ips:
                    if cr.service == constants.TELNET.SERVICE_NAME:
                        try:
                            forward_port = s.emulation_env_config.get_port_forward_port()
                            agent_transport = proxy_conn.conn.get_transport()
                            tunnel_thread = ForwardTunnelThread(local_port=forward_port,
                                                                remote_host=ip, remote_port=cr.port,
                                                                transport=agent_transport)
                            tunnel_thread.start()
                            target_conn = telnetlib.Telnet(host=constants.TELNET.LOCALHOST, port=forward_port,
                                                           timeout=3)
                            target_conn.read_until(constants.TELNET.LOGIN_PROMPT, timeout=3)
                            target_conn.write((cr.username + "\n").encode())
                            target_conn.read_until(constants.TELNET.PASSWORD_PROMPT, timeout=3)
                            target_conn.write((cr.pw + "\n").encode())
                            response = target_conn.read_until(constants.TELNET.PROMPT, timeout=3)
                            response_1 = response.decode()
                            if constants.TELNET.INCORRECT_LOGIN not in response_1 and response_1 != "":
                                connected = True
                                credentials.append(cr)
                                target_connections.append(target_conn)
                                proxies.append(proxy_conn)
                                tunnel_threads.append(tunnel_thread)
                                forward_ports.append(forward_port)
                                ports.append(cr.port)
                                ip = ip
                                non_failed_credentials.append(cr)
                            break
                        except Exception as e:
                            Logger.__call__().get_logger().warning(f"telnet exception:{str(e)}, {repr(e)}")
                            if s.attacker_obs_state is not None:
                                Logger.__call__().get_logger().warning(
                                    f"Target ip in agent reachable: {s.attacker_obs_state.agent_reachable}")
                                Logger.__call__().get_logger().warning(
                                    f"Agent reachable:{s.attacker_obs_state.agent_reachable}")
                    else:
                        non_failed_credentials.append(cr)
                if connected:
                    break
            if connected:
                break
        end = time.time()
        connection_setup_dto = ConnectionSetupDTO(
            connected=connected, credentials=credentials, target_connections=target_connections,
            proxies=proxies, non_failed_credentials=non_failed_credentials, ip=ip,
            ports=ports, forward_ports=forward_ports, interactive_shells=interactive_shells,
            tunnel_threads=tunnel_threads, total_time=end - start)
        return connection_setup_dto

    @staticmethod
    def _telnet_finalize_connection(target_machine: EmulationAttackerMachineObservationState, i: int,
                                    connection_setup_dto: ConnectionSetupDTO) -> Tuple[bool, float]:
        """
        Helper function for finalizing a Telnet connection to a target machine and creating the DTO

        :param target_machine: the target machine to connect to
        :param i: current index
        :param connection_setup_dto: DTO with information about the connection setup
        :return: boolean whether the connection has root privileges or not, cost
        """
        start = time.time()
        root = False
        for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_CHECK_ROOT):
            connection_setup_dto.target_connections[i].write("sudo -l\n".encode())
            response = connection_setup_dto.target_connections[i].read_until(constants.TELNET.PROMPT, timeout=3)
            root = False
            if f"{connection_setup_dto.credentials[i].username} may not run sudo" not in response.decode("utf-8") \
                    and ("(ALL) NOPASSWD: ALL" in response.decode("utf-8") or
                         "(ALL : ALL) ALL" in response.decode("utf-8")):
                root = True
                break
            else:
                time.sleep(constants.ENV_CONSTANTS.SLEEP_RETRY)
        connection_dto = EmulationConnectionObservationState(conn=connection_setup_dto.target_connections[i],
                                                             credential=connection_setup_dto.credentials[i],
                                                             root=root,
                                                             service=constants.TELNET.SERVICE_NAME,
                                                             tunnel_thread=connection_setup_dto.tunnel_threads[i],
                                                             tunnel_port=connection_setup_dto.forward_ports[i],
                                                             port=connection_setup_dto.ports[i],
                                                             proxy=connection_setup_dto.proxies[i],
                                                             ip=connection_setup_dto.ip)
        target_machine.telnet_connections.append(connection_dto)
        end = time.time()
        total_time = end - start
        return root, total_time

    @staticmethod
    def _ftp_setup_connection(a: EmulationAttackerAction,
                              credentials: List[Credential],
                              proxy_connections: List[EmulationConnectionObservationState],
                              s: EmulationEnvState) -> ConnectionSetupDTO:
        """
        Helper function for setting up a FTP connection

        :param a: the action of the connection
        :param credentials: list of credentials to try
        :param proxy_connections: proxy connections
        :param s: env state
        :return: a DTO with connection setup information
        """
        target_connections: List[Any] = []
        non_failed_credentials: List[Credential] = []
        proxies: List[EmulationConnectionObservationState] = []
        ports: List[int] = []
        forward_ports: List[int] = []
        interactive_shells: List[Any] = []
        tunnel_threads: List[ForwardTunnelThread] = []
        connected: bool = False
        ip: str = ""
        start = time.time()
        for proxy_conn in proxy_connections:
            if proxy_conn.ip != s.emulation_env_config.containers_config.agent_ip and proxy_conn:
                m = None
                if proxy_conn.ip is not None:
                    m = s.get_attacker_machine(proxy_conn.ip)
                if m is None or m.ips == a.ips:
                    continue
                no_match = True
                for ip in a.ips:
                    if ip in m.reachable:
                        no_match = False
                if no_match:
                    continue
            else:
                if s.attacker_obs_state is None or s.attacker_obs_state.agent_reachable is None:
                    raise ValueError("EmulationAttackerObservationState or agent_reachable is None")
                if not a.ips_match(list(s.attacker_obs_state.agent_reachable)):
                    continue
            for cr in credentials:
                for ip in a.ips:
                    if cr.service == constants.FTP.SERVICE_NAME:
                        try:
                            forward_port = s.emulation_env_config.get_port_forward_port()
                            agent_transport = proxy_conn.conn.get_transport()
                            tunnel_thread = ForwardTunnelThread(local_port=forward_port,
                                                                remote_host=ip, remote_port=cr.port,
                                                                transport=agent_transport)
                            tunnel_thread.start()
                            target_conn = FTP()
                            target_conn.connect(host=constants.FTP.LOCALHOST, port=forward_port, timeout=5)
                            login_result = target_conn.login(cr.username, cr.pw)
                            if constants.FTP.INCORRECT_LOGIN not in login_result:
                                connected = True
                                credentials.append(cr)
                                target_connections.append(target_conn)
                                proxies.append(proxy_conn)
                                tunnel_threads.append(tunnel_thread)
                                forward_ports.append(forward_port)
                                ports.append(cr.port)
                                ip = ip
                                # Create LFTP connection too to be able to search file system
                                shell = proxy_conn.conn.invoke_shell()
                                # clear output
                                if shell.recv_ready():
                                    shell.recv(constants.COMMON.DEFAULT_RECV_SIZE)
                                shell.send(constants.FTP.LFTP_PREFIX + cr.username + ":" + cr.pw + "@" + ip + "\n")
                                time.sleep(0.5)
                                # clear output
                                if shell.recv_ready():
                                    shell.recv(constants.COMMON.DEFAULT_RECV_SIZE)
                                interactive_shells.append(shell)
                                non_failed_credentials.append(cr)
                                break
                        except Exception as e:
                            if s.attacker_obs_state is None or s.attacker_obs_state.agent_reachable is None:
                                raise ValueError("Could not obtain EmulationAttackerObservationState")
                            Logger.__call__().get_logger().warning(f"FTP exception: {str(e)}, {repr(e)}")
                            Logger.__call__().get_logger().warning(
                                f"Target ip in agent reachable "
                                f"{a.ips_match(list(s.attacker_obs_state.agent_reachable))}")
                            Logger.__call__().get_logger().warning(f"Agent reachable: "
                                                                   f"{s.attacker_obs_state.agent_reachable}")
                    else:
                        non_failed_credentials.append(cr)
                if connected:
                    break
            if connected:
                break
        end = time.time()
        connection_setup_dto = ConnectionSetupDTO(
            connected=connected, credentials=credentials, target_connections=target_connections,
            proxies=proxies, non_failed_credentials=non_failed_credentials, ip=ip,
            ports=ports, forward_ports=forward_ports, interactive_shells=interactive_shells,
            tunnel_threads=tunnel_threads, total_time=end - start)
        return connection_setup_dto

    @staticmethod
    def _ftp_finalize_connection(target_machine: EmulationAttackerMachineObservationState, i: int,
                                 connection_setup_dto: ConnectionSetupDTO) -> Tuple[bool, float]:
        """
        Helper function for creating the connection DTO for FTP

        :param target_machine: the target machine to connect to
        :param i: current index
        :param connection_setup_dto: DTO with information about the connection setup
        :return: boolean, whether the connection has root privileges, cost
        """
        root = False
        connection_dto = EmulationConnectionObservationState(
            conn=connection_setup_dto.target_connections[i], credential=connection_setup_dto.credentials[i],
            root=root,
            service=constants.FTP.SERVICE_NAME, tunnel_thread=connection_setup_dto.tunnel_threads[i],
            tunnel_port=connection_setup_dto.forward_ports[i], port=connection_setup_dto.ports[i],
            interactive_shell=connection_setup_dto.interactive_shells[i], ip=connection_setup_dto.ip,
            proxy=connection_setup_dto.proxies[i])
        target_machine.ftp_connections.append(connection_dto)
        return root, 0

    @staticmethod
    def find_jump_host_connection(ip, s: EmulationEnvState) -> EmulationConnectionObservationState:
        """
        Utility function for finding a jump-host from the set of compromised machines to reach a target IP

        :param ip: the ip to reach
        :param s: the current state
        :return: a connection DTO
        """
        if s.attacker_obs_state is None or s.attacker_obs_state.agent_reachable is None:
            raise ValueError("EmulationAttackerObservationState or agent_reachable is None")
        if ip in s.attacker_obs_state.agent_reachable:
            cr = Credential(
                username=constants.AGENT.USER, pw=constants.AGENT.PW, root=True,
                protocol=TransportProtocol.TCP, service=constants.SSH.SERVICE_NAME, port=constants.SSH.DEFAULT_PORT)
            c = EmulationConnectionObservationState(
                conn=s.emulation_env_config.get_hacker_connection(), credential=cr,
                root=True, port=constants.SSH.DEFAULT_PORT, service=constants.SSH.SERVICE_NAME,
                proxy=None, ip=s.emulation_env_config.containers_config.agent_ip)
            return c
        s.attacker_obs_state.sort_machines()
        err_str = ""
        for m in s.attacker_obs_state.machines:
            err_str = err_str + (f"ips: {m.ips}, logged in: {m.logged_in}, tools installed: {m.tools_installed}, "
                                 f"reachable: {m.reachable} \n")
            if m.logged_in and m.tools_installed and m.backdoor_installed and ip in m.reachable:

                # Start with ssh connections
                ssh_connections_sorted_by_root = sorted(
                    m.ssh_connections,
                    key=lambda x: (constants.SSH_BACKDOOR.BACKDOOR_PREFIX in x.credential.username, x.root,
                                   x.credential.username),
                    reverse=True)

                for c in ssh_connections_sorted_by_root:
                    alive = ConnectionUtil.test_connection(c)
                    if alive:
                        return c
        raise ValueError(f"No JumpHost found for ip: {ip}, topology: {err_str}")

    @staticmethod
    def test_connection(c: EmulationConnectionObservationState) -> bool:
        """
        Utility function for testing if a connection is alive or not

        :param c: the connection to thest
        :return: True if the connection is alive, otherwise false
        """
        cmd = constants.AUXILLARY_COMMANDS.WHOAMI
        outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=c.conn)
        if outdata is not None and outdata.decode("utf-8") != "":
            return True
        else:
            return False

    @staticmethod
    def reconnect_ssh(c: EmulationConnectionObservationState) -> EmulationConnectionObservationState:
        """
        Reconnects the given SSH connection if it has died for some reason

        :param c: the connection to reconnect
        :return: the new connection
        """
        if c.conn is None:
            raise ValueError("Connection failed")
        if c.proxy is None:
            c.conn = paramiko.SSHClient()
            c.conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            c.conn.connect(c.ip, username=c.credential.username, password=c.credential.pw)
            transport = c.conn.get_transport()
            if transport is not None:
                transport.set_keepalive(5)
            else:
                raise ValueError("Connection failed")
        else:
            proxy = c.proxy
            if proxy.conn.get_transport() is None or not proxy.conn.get_transport().is_active():
                proxy = ConnectionUtil.reconnect_ssh(c=proxy)

            agent_addr = (proxy.ip, c.credential.port)
            target_addr = (c.ip, c.credential.port)
            agent_transport = proxy.conn.get_transport()
            relay_channel = agent_transport.open_channel(constants.SSH.DIRECT_CHANNEL, target_addr,
                                                         agent_addr)
            c.conn = paramiko.SSHClient()
            c.conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            c.conn.connect(c.ip, username=c.credential.username, password=c.credential.pw, sock=relay_channel)
            transport = c.conn.get_transport()
            if transport is not None:
                transport.set_keepalive(5)
            else:
                raise ValueError("Connection failed")
            c.proxy = proxy
        return c

    @staticmethod
    def reconnect_telnet(c: EmulationConnectionObservationState, forward_port: int = 9000) \
            -> EmulationConnectionObservationState:
        """
        Reconnects the given Telnet connection if it has died for some reason

        :param c: the connection to reconnect
        :param forward_port: the port for port-forwarding
        :return: the new connection
        """
        if c.proxy is None:
            raise ValueError("Could not obtain proxy")
        if c.proxy.conn.get_transport() is None or not c.proxy.conn.get_transport().is_active():
            proxy = ConnectionUtil.reconnect_ssh(c=c.proxy)
            c.proxy = proxy
        for i in range(constants.ENV_CONSTANTS.NUM_RETRIES):
            try:
                agent_transport = c.proxy.conn.get_transport()
                if c.ip is None:
                    raise ValueError("The EmulationConnectionObservationState ip is None")
                tunnel_thread = ForwardTunnelThread(local_port=forward_port,
                                                    remote_host=c.ip, remote_port=constants.TELNET.DEFAULT_PORT,
                                                    transport=agent_transport)
                tunnel_thread.start()
                target_conn = telnetlib.Telnet(host=constants.TELNET.LOCALHOST, port=forward_port, timeout=3)
                target_conn.read_until(constants.TELNET.LOGIN_PROMPT, timeout=3)
                target_conn.write((c.credential.username + "\n").encode())
                target_conn.read_until(constants.TELNET.PASSWORD_PROMPT, timeout=3)
                target_conn.write((c.credential.pw + "\n").encode())
                response = target_conn.read_until(constants.TELNET.PROMPT, timeout=3)
                response_1 = response.decode()
                if constants.TELNET.INCORRECT_LOGIN not in response_1 and response_1 != "":
                    c.conn = target_conn
            except Exception as e:
                Logger.__call__().get_logger().warning(f"telnet exception:{str(e)}, {repr(e)}")
        return c



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class Flag(JSONSerializable):
    """
    Class that represents a flag in the environment
    """

    def __init__(self, name: str, dir: str, id: int, path: str, requires_root: bool = False, score: int = 1):
        """
        Initializes the DTO

        :param name: the name of the flag
        :param id: the id of the flag
        :param dir: the directory
        :param path: the path of the flag
        :param requires_root: whether the flag requires root or not
        :param score: the score of the flag
        """
        self.name = name
        self.id = id
        self.path = path
        self.requires_root = requires_root
        self.score = score
        self.dir = dir

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return "name:{}, id:{}, path:{}, requires_root:{}, score:{}, dir:{}".format(
            self.name, self.id, self.path, self.requires_root, self.score, self.dir
        )

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Flag":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = Flag(
            name=d["name"], dir=d["dir"], id=d["id"], path=d["path"], requires_root=d["requires_root"],
            score=d["score"]
        )
        return obj

    def to_dict(self) -> Dict[str, Union[str, int, bool]]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, int, bool]] = {}
        d["name"] = self.name
        d["dir"] = self.dir
        d["id"] = self.id
        d["path"] = self.path
        d["requires_root"] = self.requires_root
        d["score"] = self.score
        return d

    def __hash__(self) -> int:
        """
        :return: a hash representation of the object
        """
        return hash(self.id)

    def __eq__(self, other) -> bool:
        """
        Tests equality with another flag

        :param other: the flag to compare with
        :return: True if equal otherwise False
        """
        if not isinstance(other, Flag):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.id == other.id and self.name == other.name and self.path == other.path

    @staticmethod
    def from_json_file(json_file_path: str) -> "Flag":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return Flag.from_dict(json.loads(json_str))

    def copy(self) -> "Flag":
        """
        :return: a copy of the DTO
        """
        return Flag.from_dict(self.to_dict())

    @staticmethod
    def schema() -> "Flag":
        """
        :return: get the schema of the DTO
        """
        return Flag(name="", dir="", id=-1, path="", requires_root=True, score=-1)
